#!/usr/bin/python3

"""Import packages"""
import argparse
from argparse import RawTextHelpFormatter
import sys
from asf_metadata.metadata.asf_metadata import iso_template2lists, add_dem_lists, \
  iso_xml_structure, meta_xml_file, iso_dictionary_structure, meta_json_file


# Wrapper for the command line version
def generate_iso_template(excel_file, dem_file, iso_base):
    """Generate ISO template"""

    ### Extract ISO template from Excel spreadsheet
    print('\nExtracting ISO metadata template from Excel spreadsheet ...')
    (iso_template, iso_params, iso_values) = \
        iso_template2lists(excel_file, 'ISO Metadata Structure')
    if dem_file:
        (dem_template, dem_params, dem_values) = \
        iso_template2lists(dem_file, 'ISO Metadata Structure')
        (iso_template, iso_params, iso_values) = add_dem_lists(iso_template, \
        iso_params, iso_values, dem_template, dem_params, dem_values)

    ### Write ISO template to files
    print('Writing ISO metadata template to files ...')
    iso_structure = iso_xml_structure(excel_file, iso_template, iso_params,
        iso_values, True)
    meta_xml_file(iso_structure, iso_base + '.xml')
    iso_structure = iso_dictionary_structure(excel_file, dem_file, \
        iso_template, iso_params, iso_values)
    meta_json_file(iso_structure, iso_base + '.json')


if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='generate_iso_template.py',
        description='Generate ISO template file from Excel spreadsheet',
        formatter_class=RawTextHelpFormatter)
    parser.add_argument('excel_file',
        help='name of the Excel template spreadsheet')
    parser.add_argument('-dem', default=None,
        help='name of DEM template spreadsheet')
    parser.add_argument('iso_base',
        help='basename of the ISO metadata template file')
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()

    generate_iso_template(args.excel_file, args.dem, args.iso_base)



#!/usr/bin/python3

"""Import packages"""
from itertools import groupby
from operator import itemgetter
import os
import configparser
import pandas as pd
from sqlalchemy import create_engine
import requests
import remotezip


def split_element(key, value, out):
    """Split element"""

    key, *rest = key.split('/', 1)
    if rest:
        split_element(rest[0], value, out.setdefault(key, {}))
    else:
        out[key] = value


def nest_dict(flat_dict):
    """Build nested dictionary"""

    result = {}
    for key, value in flat_dict.items():
        split_element(key, value, result)

    return result


def is_float(value):
    """Is value float?"""

    try:
        float(value)
        return True
    except ValueError:
        return False


def is_int(value):
    """Is value integer?"""

    try:
        int(value)
        return True
    except ValueError:
        return False


def get_value(value_str):
    """Get value"""

    if is_int(value_str):
        out_value = int(value_str)
    elif is_float(value_str):
        out_value = float(value_str)
    else:
        out_value = value_str.strip()
    return out_value


def rreplace(string, find, replace):
    """Replace from right"""

    reversed_string = string[::-1]
    replaced = reversed_string.replace(find[::-1], replace[::-1], 1)
    return replaced[::-1]


def unique_list(str_list):
    """Build unique list"""

    out_list = []
    for element in str_list:
        if element not in out_list:
            out_list.append(element)
    return out_list


def meta2dict(meta_file):
    """Convert metadata to dictionary"""

    ### Extract parameters and values from metadata file
    meta_params = []
    meta_values = []
    i = 0
    section = ''
    with open(meta_file, encoding='utf-8') as file:
        lines = [x.strip() for x in file.readlines()]
    for line in lines:
        if '{' in line:
            group = line.split(' {')[0].lstrip()
            section = group
            if section == 'vector':
                i += 1
        elif ':' in line:
            param = line.split(':')[0].lstrip()
            if section == 'vector':
                meta_params.append(f'state/vector[{i}]/{param}')
            elif section == 'param':
                meta_params.append(f'projection/param/{param}')
            elif section == 'utm':
                meta_params.append(f'projection/param/utm/{param}')
            elif section == 'band_stats':
                meta_params.append(f'statistics/band_stats/{param}')
            else:
                if param == 'meta_version':
                    meta_params.append(param)
                else:
                    meta_params.append(section + '/' + param)
            value = line.split(': ')[1].split('#')[0].rstrip()
            if is_int(value):
                meta_values.append(int(value))
            elif is_float(value):
                meta_values.append(float(value))
            else:
                meta_values.append(value)
        elif '}' in line:
            section = ''
        else:
            continue

    ### Establish flattened dictionary
    flat_dict = dict(zip(meta_params, meta_values))

    ### Convert flattened into nested dictionary
    meta_dict = nest_dict(flat_dict)

    return meta_dict


def merge_dict(first, second):
    """Merge dictionary"""

    key = None
    if first is None or isinstance(first, (str, int, float)):
        first = second
    elif isinstance(first, list):
        if isinstance(second, list):
            if len(first[0]) == len(second[0]):
                first[0] = merge_dict(first[0], second[0])
            elif next(iter(second[0])) not in first[len(first)-1].keys():
                first[len(first)-1] = \
                    merge_dict(first[len(first)-1], second[0])
            else:
                first.extend(second)
        else:
            first.append(second)
    elif isinstance(first, dict):
        if isinstance(second, dict):
            for key in second:
                if key in first:
                    first[key] = merge_dict(first[key], second[key])
                else:
                    first[key] = second[key]

    return first


def get_consecutive_list(meta_list):
    """Get consecutive list"""

    for _, group in groupby(enumerate(meta_list), lambda x: x[0]-x[1]):
        yield list(map(itemgetter(1), group))


def get_level_params_list(meta_template, meta_attributes, level):
    """Get level parameter list"""

    level_count = 0
    row_count = len(meta_template)
    for i in range(row_count):
        meta_element = meta_template[i].split('/')
        level_count = len(meta_element) \
            if len(meta_element) > level_count else level_count

    dataframe = pd.DataFrame(index=range(row_count), columns=range(level_count))
    for i in range(row_count):
        meta_element = meta_template[i].split('/')
        for k in range(len(meta_element)):
            dataframe.at[i,k] = meta_element[k]
        if meta_attributes:
            if isinstance(meta_attributes[i], str) and level == k+1:
                dataframe.at[i,k+1] = meta_attributes[i]
    row_index_list = list(dataframe.count(axis=1) - 1)
    indices = [i for i, x in enumerate(row_index_list) if x == level]
    indices = list(dataframe.filter(items=indices, axis=0).index)
    level_params = []
    for index in indices:
        param = '/'.join(list(dataframe.loc[index].dropna()))
        if param.endswith(']'):
            level_params.append(param.rsplit('[',1)[0])
        else:
            level_params.append(param)

    return level_params


def get_params_dataframe(meta_params):
    """Get parameter dataframe"""

    level_count = 0
    row_count = len(meta_params)
    for i in range(row_count):
        meta_element = meta_params[i].split('/')
        level_count = \
            len(meta_element) if len(meta_element) > level_count else level_count

    dataframe = pd.DataFrame(index=range(row_count),
        columns=range(level_count-1), dtype=int)
    for i in range(row_count):
        meta_element = meta_params[i].split('/')
        for k in range(level_count):
            if k < len(meta_element):
                if meta_element[k].endswith(']'):
                    dataframe.at[i,k] = int(meta_element[k].split('[')[1][:-1])
                else:
                    dataframe.at[i,k] = 0
            else:
                dataframe.at[i,k] = None

    return dataframe


def get_dict_list_params(dict_list, old_level, new_level):
    """Get dictionary list parameters"""

    dict_parents = []
    dict_params = []
    for dict_element in dict_list:
        dict_parents.append(dict_element['path'] \
            .rsplit('/',old_level-new_level)[0])
        dict_params.append(dict_element['path'])

    return (dict_parents, dict_params)


def params2dict_list(dataframe, meta_params, meta_values, level):
    """Convert parameters to dictionary list"""

    dict_list = []
    row_index = list(dataframe.count(axis=1) - 1)
    col_index = list(dataframe.sum(axis=0))
    for i in range(len(col_index)):
        if col_index[i] > 0:
            row_index.append(i)
    row_index = pd.Series(row_index)
    dataframe_column = dataframe[level].dropna()
    row_level_index = row_index[row_index==level]
    dataframe_column = \
        dataframe_column.filter(items=list(row_level_index.index))
    dataframe_column_length = len(dataframe_column)
    if dataframe_column_length > 0:
        min_list = int(dataframe_column.min())
        max_list = int(dataframe_column.max())
        for list_index in range(min_list, max_list+1):
            dataframe_list = dataframe_column[dataframe_column == list_index]
            dataframe_index_list = \
                list(get_consecutive_list(dataframe_list.index.tolist()))
            for dataframe_indices in dataframe_index_list:
                for dataframe_index in dataframe_indices:
                    meta_element = meta_params[dataframe_index].split('/')
                    path = meta_params[dataframe_index]
                    parent = {}
                    parent['level'] = len(path.split('/'))-1
                    parent['path'] = path
                    parent['type'] = 'value'
                    parent['key'] = meta_element[level]
                    parent['list'] = list_index
                    parent['dict'] = meta_values[dataframe_index]
                    dict_list.append(parent)

    return dict_list


def upgrade_dictionary2level(old_dict_list, old_level, new_level):
    """Upgrade dictionary to level"""

    new_dict_list = []
    (dict_parents, _) = \
        get_dict_list_params(old_dict_list, old_level, new_level)
    parents = list(dict.fromkeys(dict_parents))

    for parent in parents:
        indices = [i for i, x in enumerate(dict_parents) if x == parent]
        for index in indices:
            parent_dict = {}
            level_diff = old_dict_list[index]['level'] - new_level
            path = old_dict_list[index]['path']
            old_dict = old_dict_list[index]['dict']
            for _ in range(level_diff):
                key = path.rsplit('/',1)[1]
                if ('Code' in key) and (key != 'postalCode'):
                    value = old_dict
                    old_dict = {}
                    if 'EOS' in key:
                        code_url = 'http://earthdata.nasa.gov/metadata/' \
                            'resources/Codelist.xml#' + key
                    else:
                        code_url = 'https://cdn.earthdata.nasa.gov/iso/' \
                            'resources/Codelist/gmxCodelists.xml#' + key
                    old_dict['@codeList'] = code_url
                    old_dict['@codeListValue'] = value
                    old_dict['value'] = value
                new_dict = {}
                new_dict[key.split('[')[0]] = old_dict
                path = path.rsplit('/',1)[0]
                old_dict = new_dict
            parent_dict['level'] = new_level
            if path.endswith(']'):
                list_index = int(path.rsplit('[',1)[1][:-1])
                path = path.rsplit('[',1)[0]
                parent_dict['path'] = path
                parent_dict['key'] = path.rsplit('/',1)[1]
                parent_dict['list'] = list_index
                parent_dict['type'] = 'list'
            else:
                if isinstance(new_dict, dict):
                    parent_dict['path'] = path
                elif isinstance(new_dict, (int, float, str)):
                    parent_dict['path'] = path.rsplit('/',1)[0]
                if '/' in path:
                    parent_dict['key'] = path.rsplit('/',1)[1]
                else:
                    parent_dict['key'] = path
                parent_dict['list'] = 0
                parent_dict['type'] = 'single'
                parent_dict['dict'] = new_dict
            if level_diff > 0:
                parent_dict['dict'] = new_dict
            else:
                parent_dict['dict'] = old_dict_list[index]['dict']
            new_dict_list.append(parent_dict)

    return new_dict_list


def merge_dictionary_params(old_dict_list, level_params, old_level, new_level):
    """Merge ditionary parameters"""

    merged_dict_list = []
    (_, dict_params) = \
        get_dict_list_params(old_dict_list, old_level, new_level)
    params = list(dict.fromkeys(dict_params))
    params = [i for i, group in groupby(level_params)]

    for param in params:
        indices = [i for i, x in enumerate(dict_params) if x == param]
        list_indices = []
        for index in indices:
            list_indices.append(old_dict_list[index]['list'])
        if len(list_indices) > 0:
            max_index = max(list_indices)
        else:
            max_index = 0

        ### list case
        if max_index > 0:
            dict_indices = [[] for i in range(max_index)]
            for index in indices:
                dict_indices[old_dict_list[index]['list']-1].append(index)
            num_indices = len(dict_indices[0])
            combine_dict = []
            for dict_index in dict_indices:
                old_dict = old_dict_list[dict_index[0]]
                if num_indices > 1:
                    list_dict = {}
                    for list_index in dict_index:
                        old_dict = old_dict_list[list_index]
                        key, value = next(iter(old_dict['dict'].items()))
                        list_dict[key] = value
                    combine_dict.append(list_dict)
                else:
                    combine_dict.append(old_dict['dict'])

        ### single value case
        else:
            combine_dict = {}
            dict_type = None
            for index in indices:
                old_dict = old_dict_list[index]
                if isinstance(old_dict['dict'], dict):
                    key, value = next(iter(old_dict['dict'].items()))
                    combine_dict[key] = value
                    dict_type = 'dict'
                elif isinstance(old_dict['dict'], (float, int, str)):
                    if dict_type == 'dict':
                        combine_dict['value'] = old_dict['dict']
                    else:
                        combine_dict = old_dict['dict']

        merged_dict = {}
        merged_dict['level'] = old_dict['level']
        if old_dict['key'] == 'type' and \
            'acquisitionInformation' in old_dict['path']:
            merged_dict['path'] = param
            merged_dict['key'] = old_dict['key']
        elif old_dict['key'] in ['id','href','nilReason','type','uom'] and \
            'additionalAttribute' not in old_dict['path']:
            merged_dict['path'] = rreplace(param, '/','/@')
            merged_dict['key'] = '@' + old_dict['key']
        else:
            merged_dict['path'] = param
            merged_dict['key'] = old_dict['key']
        merged_dict['list'] = old_dict['list']
        merged_dict['type'] = 'single'
        merged_dict['dict'] = combine_dict
        merged_dict_list.append(merged_dict)

    return merged_dict_list


def create_postgis_engine(config_file):
    """Create PostGIS engine"""

    config = configparser.ConfigParser()
    config.read(config_file)
    connection_string = \
        "postgresql://" + config.get('postgres', 'user') + \
        ":'" + config.get('postgres', 'pass') + "'" + \
        "@" + config.get('postgres', 'host') + ":5432" + \
        "/" + config.get('postgres', 'db')
    engine = create_engine(connection_string)

    return engine


def access_remote_zip(zip_file):
    """Access remote zip file"""

    ### Set up cookie jar
    login_url = 'https://urs.earthdata.nasa.gov/oauth/authorize?client_id=' \
        'BO_n7nTIlMljdvU6kRRB3g&response_type=code&redirect_uri=' \
        'https://auth.asf.alaska.edu/login&app_type=401'
    jar = requests.cookies.RequestsCookieJar()
    username = os.getenv('EDL_USER')
    password = os.getenv('EDL_PASS')
    login = requests.get(login_url, auth=(username, password))
    for row in login.history:
        jar.update(row.cookies.get_dict())

    zip_filehandle = remotezip.RemoteZip(zip_file, cookies=jar)

    return zip_filehandle



#!/usr/bin/python3

import argparse
from argparse import RawTextHelpFormatter
import sys
from asf_metadata.metadata.asf_metadata import iso_template2lists, add_dem_lists, \
  iso_dictionary_structure, get_metadata_values, meta_json_file


def iso_metadata_xml2json(excelFile, demFile, xmlFile, jsonFile):

  ### Read ISO template
  (isoTemplate, isoParams, isoValues) = \
    iso_template2lists(excelFile, 'ISO Metadata Structure')
  if demFile:
    (demTemplate, demParams, demValues) = \
      iso_template2lists(demFile, 'ISO Metadata Structure')
    (isoTemplate, isoParams, isoValues) = add_dem_lists(isoTemplate, \
      isoParams, isoValues, demTemplate, demParams, demValues)

  ### Extract values from ISO XML
  isoProdValues = get_metadata_values(xmlFile, isoParams)

  ### Write ISO metadata to JSON
  isoStructure = iso_dictionary_structure(excelFile, demFile, isoTemplate, \
    isoParams, isoProdValues)
  meta_json_file(isoStructure, jsonFile)


if __name__ == '__main__':

  parser = argparse.ArgumentParser(prog='iso_metadata_xml2json.py',
    description='Convert ISO metadata from XML to JSON format',
    formatter_class=RawTextHelpFormatter)
  parser.add_argument('excelFile', help='name of the Excel template spreadsheet')
  parser.add_argument('-dem', default=None,
    help='name of DEM template spreadsheet')
  parser.add_argument('xmlFile', help='name of the ISO metadata XML file')
  parser.add_argument('jsonFile', help='name of the ISO metadata JSON file')
  if len(sys.argv) == 1:
    parser.print_help()
    sys.exit(1)
  args = parser.parse_args()

  iso_metadata_xml2json(args.excelFile, args.dem, args.xmlFile, args.jsonFile)



#!/usr/bin/python3

"""Import packages"""
import argparse
from argparse import RawTextHelpFormatter
import os
import sys
from asf_metadata.metadata.asf_metadata import iso_template2lists, add_dem_lists, \
  generate_product_dictionary, product_dictionary2values, iso_xml_structure, \
  meta_xml_file, iso_dictionary_structure, meta_json_file, \
  clean_json_structure, clean_xml_structure


def generate_iso_metadata(meta_file, log_file, iso_base, product_type,
    data_type, dem_type):
    """Generate ISO metadata"""

    ### Work out the location of the templates and their files
    template_path = \
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')
    data_template_file = f"{data_type}_{product_type}_metadata.xlsx"
    excel_file = os.path.join(template_path, data_type, data_template_file)
    print(f'\nData template: {excel_file}')
    if dem_type:
        dem_template_file = f"dem_{dem_type}_metadata.xlsx"
        dem_file = os.path.join(template_path, 'dem', dem_template_file)
        print(f'DEM template: {dem_file}')
    else:
        dem_file = None

    ### Read ISO template
    (iso_template, iso_params, iso_values) = \
        iso_template2lists(excel_file, 'ISO Metadata Structure')
    if dem_type:
        (dem_template, dem_params, dem_values) = \
            iso_template2lists(dem_file, 'ISO Metadata Structure')
        (iso_template, iso_params, iso_values) = add_dem_lists(iso_template, \
            iso_params, iso_values, dem_template, dem_params, dem_values)
    else:
        dem_params = None
        dem_template = None

    ### Generate product dictionary
    iso_prod_dict = generate_product_dictionary(product_type,
        data_type, meta_file, log_file)

    ### Map product dictionary to values
    iso_prod_values = product_dictionary2values(iso_values, iso_prod_dict)

    ### Write ISO metadata to XML
    iso_file = iso_base + '.iso.xml'
    print(f'Writing ISO metadata structure to XML file ({iso_file}) ...')
    iso_structure = iso_xml_structure(excel_file, iso_template, iso_params, \
        iso_prod_values, True)
    iso_structure = clean_xml_structure(iso_file)
    meta_xml_file(iso_structure, iso_file)

    ### Write ISO metadata to JSON
    iso_file = iso_base + '.iso.json'
    print(f'Writing ISO metadata structure to JSON file ({iso_file}) ...')
    iso_structure = iso_dictionary_structure(excel_file, dem_file, \
        iso_template, iso_params, iso_prod_values)
    iso_structure = clean_json_structure(iso_structure)
    meta_json_file(iso_structure, iso_file)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='generate_iso_metadata.py',
        description='Generate ISO metadata file from Excel spreadsheet',
        formatter_class=RawTextHelpFormatter)
    parser.add_argument('meta_file', help='file path of product metadata ' \
        'such as a "manifest.safe" or leader file')
    parser.add_argument('log_file', help='name of RTC processing log file')
    parser.add_argument('iso_base', help='basename of the ISO XML metadata file')
    parser.add_argument('-product', default='gamma_rtc',
        help='name of the product type (default: gamma_rtc)\n' \
            'choices: ["gamma_rtc"]')
    parser.add_argument('-data', default='sentinel',
        help='name of the data source (default: sentinel)\n' \
            'choices: ["sentinel"]')
    parser.add_argument('-dem', default=None,
        help='DEM type used for terrain correction\n' \
            'choices: ["copernicus"]')
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()

    generate_iso_metadata(args.meta_file, args.log_file, args.iso_base,
        args.product, args.data, args.dem)



# to change color of output to user:
class colors:
    WARNING = '\033[93m'
    ERROR = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'

# Make sure GDAL is setup
try:
    from osgeo import ogr
except ImportError as e:
    raise ImportError(f"""
    {colors.ERROR}
    ERROR: Could not find the GDAL/OGR Python library bindings. 
    On Debian based systems you can install it with these commands:
        sudo apt install -y python3-gdal libgdal-dev
        python3 -m pip install gdal==$(gdal-config --version) --global-option=build_ext --global-option='-I/usr/include/gdal/'
    {colors.END}
    """) from e

# Import what the user is supposed to use directly
from asf_metadata.generate_iso_metadata import generate_iso_metadata
from asf_metadata.generate_iso_template import generate_iso_template
from asf_metadata.iso_metadata_xml2json import iso_metadata_xml2json



#!/usr/bin/python3

"""Import packages"""
import argparse
from argparse import RawTextHelpFormatter
import sys
from metadata.util import create_postgis_engine, access_remote_zip
from metadata.sentinel_metadata import get_sentinel_bursts


def sentinel_bursts2db(zip_url, burst_map_file, config_file):
    """Extract Sentinel burst metadata and save in a database"""

    ### Connect to postreSQL database
    engine = create_postgis_engine(config_file)

    ### Work out access to remote zip file
    zip_handle = access_remote_zip(zip_url)
    name_list = zip_handle.namelist()
    safe_dir = name_list[0]

    ### Get Sentinel-1 file information
    print('\nSentinel-1 bursts to database ...')
    gdf = get_sentinel_bursts(safe_dir, burst_map_file, zip_handle, zip_url)
    zip_handle.close()
    gdf.to_postgis('sentinel_bursts', con=engine, if_exists='append',
      index=False)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='sentinel_bursts2db',
        description='Extract burst information from a remote granule zip ' \
            'file and save it into database',
        formatter_class=RawTextHelpFormatter)
    parser.add_argument('zip_url', metavar='<zip file URL>',
        help='URL to the remote granule zip file')
    parser.add_argument('burst_map', metavar='<burst map file>',
        help='name of the burst map GeoJSON file')
    parser.add_argument('config', metavar='<DB config file>',
        help='name of the database configuration file')
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()

    sentinel_bursts2db(args.zip_url, args.burst_map, args.config)



#!/usr/bin/python3

"""Import packages"""
import os
from datetime import datetime, timedelta
import json
import lxml.etree as et
from osgeo import gdal, osr, ogr
import pandas as pd
import geopandas as gpd
import numpy as np
import scipy.interpolate
from shapely import wkt
import requests
import xmltodict


def xml2string(in_file, zip_handle=None):
    """Convert XML to string"""

    if zip_handle is None:
        parser = et.XMLParser(remove_blank_text=True)
        doc = et.parse(in_file, parser)
        string = et.tostring(doc)
    else:
        string = zip_handle.read(in_file)

    return string


def get_latlon_extent(filename):
    """Get lat/lon extent"""

    src = gdal.Open(filename)
    ulx, xres, _, uly, _, yres  = src.GetGeoTransform()
    lrx = ulx + (src.RasterXSize * xres)
    lry = uly + (src.RasterYSize * yres)

    source = osr.SpatialReference()
    source.ImportFromWkt(src.GetProjection())
    target = osr.SpatialReference()
    target.ImportFromEPSG(4326)
    transform = osr.CoordinateTransformation(source, target)

    (lon1, lat1, _) = transform.TransformPoint(ulx, uly)
    (lon2, lat2, _) = transform.TransformPoint(lrx, uly)
    (lon3, lat3, _) = transform.TransformPoint(ulx, lry)
    (lon4, lat4, _) = transform.TransformPoint(lrx, lry)

    lat_min = min(lat1,lat2,lat3,lat4)
    lat_max = max(lat1,lat2,lat3,lat4)
    lon_min = min(lon1,lon2,lon3,lon4)
    lon_max = max(lon1,lon2,lon3,lon4)

    return lat_min, lat_max, lon_min, lon_max


def extract_metadata(safe_dir, zip_handle=None):
    """Extract metadata"""

    manifest = read_manifest_file(os.path.join(safe_dir, 'manifest.safe'),
        zip_handle)
    sentinel_files = get_sentinel_files(safe_dir, manifest)
    for sentinel_file in sentinel_files:
        if sentinel_file['type'] == 'annotation':
            if sentinel_file['polarization'] in ('VV','HH'):
                if sentinel_file['swath'] == 'IW1':
                    annotation_iw1 = \
                        read_annotation_file(sentinel_file['name'], zip_handle)
                elif sentinel_file['swath'] == 'IW2':
                    annotation_iw2 = \
                        read_annotation_file(sentinel_file['name'], zip_handle)
                elif sentinel_file['swath'] == 'IW3':
                    annotation_iw3 = \
                        read_annotation_file(sentinel_file['name'], zip_handle)

    return (manifest, annotation_iw1, annotation_iw2, annotation_iw3)


def read_manifest_file(manifest_file, zip_handle=None):
    """Read manifest file"""

    ns_xfdu = {'xfdu': 'urn:ccsds:schema:xfdu:1'}
    ns_all = dict(
        list(ns_xfdu.items())
    )

    ### Read manifest file
    doc = et.fromstring(xml2string(manifest_file, zip_handle))

    ### Pack first level into strings
    meta = {}
    meta['informationPackageMap'] = et.tostring( \
        doc.xpath('/xfdu:XFDU/informationPackageMap', namespaces=ns_all)[0])
    meta['metadataSection'] = \
        et.tostring(doc.xpath('/xfdu:XFDU/metadataSection',
            namespaces=ns_all)[0])
    meta['dataObjectSection'] = \
        et.tostring(doc.xpath('/xfdu:XFDU/dataObjectSection',
            namespaces=ns_all)[0])

    return meta


def read_annotation_file(annotation_file, zip_handle=None):
    """Read annotation file"""

    ### Read annotation file
    doc = et.fromstring(xml2string(annotation_file, zip_handle))

    ### Pack first level into strings
    meta = {}
    meta['adsHeader'] = et.tostring(doc.xpath('/product/adsHeader')[0])
    meta['qualityInformation'] = \
        et.tostring(doc.xpath('/product/qualityInformation')[0])
    meta['generalAnnotation'] = \
        et.tostring(doc.xpath('/product/generalAnnotation')[0])
    meta['imageAnnotation'] = \
        et.tostring(doc.xpath('/product/imageAnnotation')[0])
    meta['dopplerCentroid'] = \
        et.tostring(doc.xpath('/product/dopplerCentroid')[0])
    meta['antennaPattern'] = et.tostring(doc.xpath('/product/antennaPattern')[0])
    meta['swathTiming'] = et.tostring(doc.xpath('/product/swathTiming')[0])
    meta['geolocationGrid'] = \
        et.tostring(doc.xpath('/product/geolocationGrid')[0])
    meta['coordinateConversion'] = \
        et.tostring(doc.xpath('/product/coordinateConversion')[0])
    meta['swathMerging'] = et.tostring(doc.xpath('/product/swathMerging')[0])

    return meta


def annotation2spline(meta):
    """Extact spline function from annotation file"""

    ### Extract geolocation grid from metadata
    doc_grid = et.fromstring(meta['geolocationGrid'])
    number_of_points = int(doc_grid.xpath('/geolocationGrid/' \
        'geolocationGridPointList/@count')[0])

    ### Determine line/pixel indices for spline calculation
    lines = []
    pixels = []
    for point in range(number_of_points):
        xml = ('/geolocationGrid/geolocationGridPointList/' \
            'geolocationGridPoint[{0}]/line'.format(point+1))
        line = int(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
            'geolocationGridPoint[{0}]/pixel'.format(point+1))
        pixel = int(doc_grid.xpath(xml)[0].text)
        if line == 0:
            pixels.append(pixel)
        if pixel == 0:
            lines.append(line)
    line_vector = np.array(lines)
    line_count = len(lines)
    pixel_vector = np.array(pixels)
    pixel_count = len(pixels)

    ### Build latitude/longitude grid
    lat_grid = np.zeros((line_count, pixel_count), dtype=float)
    lon_grid = np.zeros((line_count, pixel_count), dtype=float)
    height_grid = np.zeros((line_count, pixel_count), dtype=float)
    for point in range(number_of_points):
        xml = ('/geolocationGrid/geolocationGridPointList/' \
        'geolocationGridPoint[{0}]/line'.format(point+1))
        line = int(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
        'geolocationGridPoint[{0}]/pixel'.format(point+1))
        pixel = int(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
        'geolocationGridPoint[{0}]/latitude'.format(point+1))
        lat = float(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
        'geolocationGridPoint[{0}]/longitude'.format(point+1))
        lon = float(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
        'geolocationGridPoint[{0}]/height'.format(point+1))
        height = float(doc_grid.xpath(xml)[0].text)
        index_line = lines.index(line)
        index_pixel = pixels.index(pixel)
        lat_grid[index_line, index_pixel] = lat
        lon_grid[index_line, index_pixel] = lon
        height_grid[index_line, index_pixel] = height

    ### Calculate spline for interpolation
    lat_spline = scipy.interpolate.RectBivariateSpline(line_vector,
        pixel_vector, lat_grid)
    lon_spline = scipy.interpolate.RectBivariateSpline(line_vector,
        pixel_vector, lon_grid)
    height_spline = scipy.interpolate.RectBivariateSpline(line_vector,
        pixel_vector, height_grid)

    return (lat_spline, lon_spline, height_spline)


def annotation2burst_location(meta, burst_map, granule_info):
    """Extract burst location info from annoation file"""

    ### Initialize geodataframe and define data types
    dataframe = pd.DataFrame()

    (lat_spline, lon_spline, height_spline) = annotation2spline(meta)

    ### Extract metadata
    speed_of_light = 299792458.0
    doc = et.fromstring(meta['adsHeader'])
    swath = doc.xpath('/adsHeader/swath')[0].text
    swath_number = int(swath[-1:])
    polarization = doc.xpath('/adsHeader/polarisation')[0].text
    stop_time = doc.xpath('/adsHeader/stopTime')[0].text

    doc_general = et.fromstring(meta['generalAnnotation'])
    range_sampling_rate = float(doc_general.xpath('/generalAnnotation/' \
        'productInformation/rangeSamplingRate')[0].text)
    radar_wavelength = speed_of_light / \
        float(doc_general.xpath('/generalAnnotation/productInformation/' \
        'radarFrequency')[0].text)
    azimuth_steering_rate = float(doc_general.xpath('/generalAnnotation/' \
        'productInformation/azimuthSteeringRate')[0].text)
    prf = float(doc_general.xpath('/generalAnnotation/' \
        'downlinkInformationList/downlinkInformation/prf')[0].text)
    azimuth_fm_rate_list_count = int(doc_general.xpath('/generalAnnotation/' \
        'azimuthFmRateList/@count')[0])
    fm_rates = []
    for rate in range(azimuth_fm_rate_list_count):
        xml = ('/generalAnnotation/azimuthFmRateList/azimuthFmRate[{0}]/' \
            'azimuthTime'.format(rate+1))
        ref_time = datetime.fromisoformat(doc_general.xpath(xml)[0].text)
        xml = ('/generalAnnotation/azimuthFmRateList/azimuthFmRate[{0}]/' \
            'azimuthFmRatePolynomial'.format(rate+1))
        poly_coeffs = \
            list(map(float, doc_general.xpath(xml)[0].text.split(' ')))
        fm_rates.append((ref_time, poly_coeffs))

    doc_image = et.fromstring(meta['imageAnnotation'])
    azimuth_time_interval = float(doc_image.xpath('/imageAnnotation/' \
        'imageInformation/azimuthTimeInterval')[0].text)
    range_pixel_spacing = float(doc_image.xpath('/imageAnnotation/' \
        'imageInformation/rangePixelSpacing')[0].text)
    range_window_type = doc_image.xpath('/imageAnnotation/' \
        'processingInformation/swathProcParamsList/swathProcParams/' \
        'rangeProcessing/windowType')[0].text
    range_window_coefficient = float(doc_image.xpath('/imageAnnotation/' \
        'processingInformation/swathProcParamsList/swathProcParams/' \
        'rangeProcessing/windowCoefficient')[0].text)
    azimuth_window_type = doc_image.xpath('/imageAnnotation/' \
        'processingInformation/swathProcParamsList/swathProcParams/' \
        'azimuthProcessing/windowType')[0].text
    azimuth_window_coefficient = float(doc_image.xpath('/imageAnnotation/' \
        'processingInformation/swathProcParamsList/swathProcParams/' \
        'azimuthProcessing/windowCoefficient')[0].text)

    doc_doppler = et.fromstring(meta['dopplerCentroid'])
    dc_estimate_list_count = int(doc_doppler.xpath('/dopplerCentroid/' \
        'dcEstimateList/@count')[0])
    dopplers = []
    for doppler in range(1,dc_estimate_list_count+1):
        xml = ('/dopplerCentroid/dcEstimateList/dcEstimate[{0}]/' \
            'azimuthTime'.format(doppler))
        ref_time = datetime.fromisoformat(doc_doppler.xpath(xml)[0].text)
        xml = ('/dopplerCentroid/dcEstimateList/dcEstimate[{0}]/' \
            'dataDcPolynomial'.format(doppler))
        poly_coeffs = \
            list(map(float, doc_doppler.xpath(xml)[0].text.split(' ')))
        dopplers.append((ref_time, poly_coeffs))

    doc_swath = et.fromstring(meta['swathTiming'])
    lines_per_burst = \
        int(doc_swath.xpath('/swathTiming/linesPerBurst')[0].text)
    samples_per_burst = \
        int(doc_swath.xpath('/swathTiming/samplesPerBurst')[0].text)
    number_of_bursts = int(doc_swath.xpath('/swathTiming/burstList/@count')[0])

    doc_grid = et.fromstring(meta['geolocationGrid'])
    number_of_points = int(doc_grid.xpath('/geolocationGrid/' \
        'geolocationGridPointList/@count')[0])

    ### Determine line/pixel indices for spline calculation
    lines = []
    pixels = []
    slant_ranges = []
    for point in range(number_of_points):
        xml = ('/geolocationGrid/geolocationGridPointList/' \
            'geolocationGridPoint[{0}]/line'.format(point+1))
        #print(doc_grid.xpath(xml))
        line = int(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
            'geolocationGridPoint[{0}]/pixel'.format(point+1))
        pixel = int(doc_grid.xpath(xml)[0].text)
        xml = ('/geolocationGrid/geolocationGridPointList/' \
            'geolocationGridPoint[{0}]/slantRangeTime'.format(point+1))
        slant_range_time = float(doc_grid.xpath(xml)[0].text)
        if line == 0:
            pixels.append(pixel)
        if pixel == 0:
            lines.append(line)
        slant_range = slant_range_time * speed_of_light / 2.0
        slant_ranges.append(slant_range)
    max_pixel = max(pixels)
    center_pixel = max_pixel / 2.0

    acq_time = []
    sensing_time = []
    anx_time = []
    for burst in range(1,number_of_bursts+1):
        xml = ('/swathTiming/burstList/burst[{0}]/azimuthAnxTime' \
            .format(burst))
        anx_time_burst = float(doc_swath.xpath(xml)[0].text)
        xml = ('/swathTiming/burstList/burst[{0}]/azimuthTime'.format(burst))
        azimuth_start_time = doc_swath.xpath(xml)[0].text
        xml = ('/swathTiming/burstList/burst[{0}]/sensingTime'.format(burst))
        sensing_start = doc_swath.xpath(xml)[0].text
        acq_time.append(datetime.fromisoformat(azimuth_start_time))
        sensing_time.append(datetime.fromisoformat(sensing_start))
        anx_time.append(anx_time_burst)
    acq_time.append(datetime.fromisoformat(stop_time))
    sensing_time.append(datetime.fromisoformat(stop_time))
    sensing_period = (lines_per_burst-1)/prf
    sensing_mid_period = sensing_period / 2.0

    ### Determine valid data and calculate burst boundaries
    for burst in range(number_of_bursts):

        first_line = lines[burst]
        last_line = lines[burst+1]
        starting_range = slant_ranges[burst]
        burst_time_period = \
            (last_line - first_line - 1) * azimuth_time_interval
        acq_period = abs(acq_time[burst+1] - acq_time[burst])
        acq_time_period = acq_period.seconds + acq_period.microseconds/1000000
        last_line += int(abs(burst_time_period - acq_time_period) / \
            azimuth_time_interval + 0.5)
        lats = np.zeros(4, dtype=float)
        lons = np.zeros(4, dtype=float)
        heights = np.zeros(4, dtype=float)
        lats[0] = lat_spline(first_line, 0)[0][0]
        lons[0] = lon_spline(first_line, 0)[0][0]
        heights[0] = height_spline(first_line, 0)[0][0]
        lats[1] = lat_spline(first_line, max_pixel)[0][0]
        lons[1] = lon_spline(first_line, max_pixel)[0][0]
        heights[1] = height_spline(first_line, max_pixel)[0][0]
        lats[2] = lat_spline(last_line, max_pixel)[0][0]
        lons[2] = lon_spline(last_line, max_pixel)[0][0]
        heights[2] = height_spline(last_line, max_pixel)[0][0]
        lats[3] = lat_spline(last_line, 0)[0][0]
        lons[3] = lon_spline(last_line, 0)[0][0]
        heights[3] = height_spline(last_line, 0)[0][0]
        center_line = first_line + (last_line - first_line) / 2.0
        terrain_height = height_spline(center_line, center_pixel)[0][0]

        xml = ('/swathTiming/burstList/burst[{0}]/azimuthAnxTime' \
            .format(burst+1))
        anx_time_burst = float(doc_swath.xpath(xml)[0].text)
        xml = ('/swathTiming/burstList/burst[{0}]/azimuthTime'.format(burst+1))
        azimuth_start_time = doc_swath.xpath(xml)[0].text
        start_time = datetime.fromisoformat(azimuth_start_time)
        stop_time = start_time + timedelta(seconds=burst_time_period)
        azimuth_stop_time = stop_time.isoformat(timespec='microseconds')
        xml = ('/swathTiming/burstList/burst[{0}]/sensingTime'.format(burst+1))
        sensing_start = doc_swath.xpath(xml)[0].text
        start_time = datetime.fromisoformat(sensing_start)
        mid_time = start_time + timedelta(seconds=sensing_mid_period)
        stop_time = start_time + timedelta(seconds=sensing_period)
        sensing_stop = stop_time.isoformat(timespec='microseconds')

        time_diff = \
            [np.abs((mid_time - val[0]).total_seconds()) for val in fm_rates]
        index = np.argmin(time_diff)
        azimuth_fm_rate_polynomial = fm_rates[index][1]

        time_diff = \
            [np.abs((mid_time - val[0]).total_seconds()) for val in dopplers]
        index = np.argmin(time_diff)
        doppler_polynomial = dopplers[index][1]

        xml = ('/swathTiming/burstList/burst[{0}]/firstValidSample' \
            .format(burst+1))
        valid_samples = \
            list(map(int, doc_swath.xpath(xml)[0].text.split(' ')))
        first_valid_sample = np.median(np.array(valid_samples)).astype(int)
        valid_indices = \
           [i for i, x in enumerate(valid_samples) if x == first_valid_sample]
        first_valid_line = int(min(valid_indices))
        last_valid_line = int(max(valid_indices))
        xml = ('/swathTiming/burstList/burst[{0}]/lastValidSample' \
            .format(burst+1))
        valid_samples = list(map(int, doc_swath.xpath(xml)[0].text.split(' ')))
        last_valid_sample = np.median(np.array(valid_samples)).astype(int)
        number_valid_lines = last_valid_line - first_valid_line - 1
        number_valid_samples = last_valid_sample - first_valid_sample - 1

        ipf = float(granule_info['ipf'])
        if ipf >= 3.4:
            xml = ('/swathTiming/burstList/burst[{0}]/burstId/@absolute' \
                .format(burst+1))
            absolute_esa_burst_id = int(doc_swath.xpath(xml)[0])
            xml = ('/swathTiming/burstList/burst[{0}]/burstId' \
                .format(burst+1))
            relative_esa_burst_id = int(doc_swath.xpath(xml)[0].text)
        else:
            absolute_esa_burst_id = -1
            relative_esa_burst_id = -1

        orbit_direction = granule_info['passDirection']
        track = granule_info['trackNumber']
        absolute_burst_id = ('%s-%06d' % (get_burst_id(burst_map,
            orbit_direction, track, swath, float(anx_time[burst])),
            granule_info['orbitNumber']))
        relative_burst_id = get_burst_id(burst_map, orbit_direction, track,
            swath, float(anx_time[burst]))
        if burst > 0:
            previous_burst_id = get_burst_id(burst_map, orbit_direction, track,
                swath, float(anx_time[burst-1]))
        else:
            previous_burst_id = None
        if burst < number_of_bursts - 1:
            next_burst_id = get_burst_id(burst_map, orbit_direction, track,
                swath, float(anx_time[burst+1]))
        else:
            next_burst_id = None

        ring = ogr.Geometry(ogr.wkbLinearRing)
        ring.AddPoint_2D(float(lons[0]), float(lats[0]))
        ring.AddPoint_2D(float(lons[1]), float(lats[1]))
        ring.AddPoint_2D(float(lons[2]), float(lats[2]))
        ring.AddPoint_2D(float(lons[3]), float(lats[3]))
        ring.AddPoint_2D(float(lons[0]), float(lats[0]))
        polygon = ogr.Geometry(ogr.wkbPolygon)
        polygon.AddGeometry(ring)

        values = pd.DataFrame()
        values.at[0,'granuleUR'] = ('%s-BURSTS-%s-%s-%02d' \
        % (granule_info['granule'], swath, polarization, burst + 1))
        values.at[0,'insertDate'] = None
        values.at[0,'productionDateTime'] = datetime.utcnow().isoformat() + 'Z'
        values.at[0,'identifier'] = ('%s-BURSTS' % granule_info['granule'])
        values.at[0,'inputGranule1'] = granule_info['granule']
        values.at[0,'beginningDateTime'] = azimuth_start_time
        values.at[0,'endingDateTime'] = azimuth_stop_time

        values.at[0,'platformShortName'] = granule_info['platform']
        values.at[0,'urlOfFrame'] = granule_info['urlOfFrame']
        values.at[0,'swath'] = swath
        values.at[0,'startLineInSwath'] = int(first_line)
        values.at[0,'endLineInSwath'] = int(last_line)
        values.at[0,'ipf'] = granule_info['ipf']

        values.at[0,'numberOfLines'] = int(lines_per_burst)
        values.at[0,'numberOfSamples'] = int(samples_per_burst)
        values.at[0,'startingRange'] = float(starting_range)
        values.at[0,'sensingStart'] = sensing_start
        values.at[0,'sensingStop'] = sensing_stop
        values.at[0,'burstStartUTC'] = azimuth_start_time
        values.at[0,'burstStopUTC'] = azimuth_stop_time
        values.at[0,'trackNumber'] = int(granule_info['trackNumber'])
        values.at[0,'frameNumber'] = int(-1)
        values.at[0,'orbitNumber'] = int(granule_info['orbitNumber'])
        values.at[0,'swathNumber'] = int(swath_number)
        values.at[0,'burstNumber'] = int(burst) + 1
        values.at[0,'passDirection'] = granule_info['passDirection']
        values.at[0,'azimuthSteeringRate'] = float(azimuth_steering_rate)
        values.at[0,'rangePixelSize'] = float(range_pixel_spacing)
        values.at[0,'rangeSamplingRate'] = float(range_sampling_rate)
        values.at[0,'azimuthTimeInterval'] = float(azimuth_time_interval)
        values.at[0,'radarWavelength'] = float(radar_wavelength)
        values.at[0,'polarization'] = polarization
        values.at[0,'terrainHeight'] = float(terrain_height)
        values.at[0,'prf'] = float(prf)
        values.at[0,'firstValidLine'] = int(first_valid_line)
        values.at[0,'numValidLines'] = int(number_valid_lines)
        values.at[0,'firstValidSample'] = int(first_valid_sample)
        values.at[0,'numValidSamples'] = int(number_valid_samples)
        values.at[0,'rangeWindowType'] = range_window_type
        values.at[0,'rangeWindowCoefficient'] = \
            float(range_window_coefficient)
        values.at[0,'azimuthWindowType'] = azimuth_window_type
        values.at[0,'azimuthWindowCoefficient'] = \
            float(azimuth_window_coefficient)
        values.at[0,'azimuthFMRate'] = \
            ','.join(map(str, azimuth_fm_rate_polynomial))
        values.at[0,'doppler'] = ','.join(map(str, doppler_polynomial))

        values.at[0,'absoluteBurstID'] = absolute_burst_id
        values.at[0,'relativeBurstID'] = relative_burst_id
        values.at[0,'previousBurstID'] = previous_burst_id
        values.at[0,'nextBurstID'] = next_burst_id
        values.at[0,'absoluteESAburstID'] = absolute_esa_burst_id
        values.at[0,'relativeESAburstID'] = relative_esa_burst_id
        values.at[0,'timeSinceAnxNode'] = anx_time_burst
        values.at[0,'ascendingNodeTime'] = granule_info['ascendingNodeTime']
        values.at[0,'hasLand'] = True
        values.at[0,'crossesDateLine'] = False
        values.at[0,'landFraction'] = 1.0

        values.at[0,'pointLongitude_1'] = float(lons[0])
        values.at[0,'pointLatitude_1'] = float(lats[0])
        values.at[0,'pointHeight_1'] = float(heights[0])
        values.at[0,'pointLongitude_2'] = float(lons[1])
        values.at[0,'pointLatitude_2'] = float(lats[1])
        values.at[0,'pointHeight_2'] = float(heights[1])
        values.at[0,'pointLongitude_3'] = float(lons[2])
        values.at[0,'pointLatitude_3'] = float(lats[2])
        values.at[0,'pointHeight_3'] = float(heights[2])
        values.at[0,'pointLongitude_4'] = float(lons[3])
        values.at[0,'pointLatitude_4'] = float(lats[3])
        values.at[0,'pointHeight_4'] = float(heights[3])
        values.at[0,'pointLongitude_5'] = float(lons[0])
        values.at[0,'pointLatitude_5'] = float(lats[0])
        values.at[0,'pointHeight_5'] = float(heights[0])
        values.at[0,'geometry'] = wkt.loads(polygon.ExportToWkt())

        dataframe = pd.concat([dataframe, values], ignore_index=True)

    return dataframe


def get_burst_id(burst_map, orbit_direction, track, swath, anx_time):
    """Get burst index"""

    burst_direction = burst_map[burst_map['orbitDir'] == orbit_direction]
    burst_track = burst_direction[burst_direction['track'] == track]
    burst_swath = pd.DataFrame(burst_track[burst_track['swath'] == swath])
    burst_swath = burst_swath.reset_index()
    delta_anx_time = 12*24*3600/175
    burst_id = 'TTT-AAAA-IIW-PP-VU'
    for i in range(len(burst_swath)):
        diff = abs(anx_time - burst_swath.at[i,'anx_time'])
        if diff < delta_anx_time:
            burst_id = burst_swath.at[i,'burst_id']
            delta_anx_time = diff

    return burst_id


def get_sentinel_bursts(safe_dir, burst_map_file, zip_handle=None, url_str=None):
    """Get Sentinel bursts"""

    ### Read burst map
    base_name = os.path.basename(burst_map_file)
    message = f'Reading burst map file ({base_name}) ...'
    print(message)
    burst_map = gpd.read_file(burst_map_file)

    ### Define namespaces for XML parsing
    ns_safe = {'safe': 'http://www.esa.int/safe/sentinel-1.0'}
    ns_s1 = {'s1': 'http://www.esa.int/safe/sentinel-1.0/sentinel-1'}
    names = dict(
        list(ns_safe.items()) +
        list(ns_s1.items())
    )

    (manifest, annotation_iw1, annotation_iw2, annotation_iw3) = \
        extract_metadata(safe_dir, zip_handle)
    granule = safe_dir[:-6]
    print(f'Extracting information for granule ({granule}) ...')

    doc = et.fromstring(manifest['metadataSection'])
    meta_object_count = len(doc.xpath('/metadataSection/metadataObject/@ID'))
    for i in range(1,meta_object_count+1):
        index = doc.xpath('/metadataSection/metadataObject[{0}]/@ID' \
            .format(i))[0]
        if index == 'processing':
            ipf = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:processing/safe:facility/' \
                'safe:software/@version'.format(i), namespaces=names)[0]
        elif index == 'platform':
            param = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:platform/safe:familyName' \
                .format(i), namespaces=names)
            platform = param[0].text
            param = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:platform/safe:number' \
                .format(i), namespaces=names)
            platform += param[0].text
        elif index == 'measurementOrbitReference':
            param = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:orbitReference/' \
                'safe:orbitNumber[@type="start"]'.format(i), namespaces=names)
            abolute_orbit_number = int(param[0].text)
            param = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:orbitReference/' \
                'safe:relativeOrbitNumber[@type="start"]'.format(i),
                namespaces=names)
            relative_orbit_number = int(param[0].text)
            param = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:orbitReference/safe:extension/' \
                's1:orbitProperties/s1:pass'.format(i), namespaces=names)
            orbit_direction = param[0].text
            param = doc.xpath('/metadataSection/metadataObject[{0}]/' \
                'metadataWrap/xmlData/safe:orbitReference/safe:extension/' \
                's1:orbitProperties/s1:ascendingNodeTime'.format(i),
                namespaces=names)
            ascending_node_time = param[0].text

    if not url_str:
        ### Determine URL (if needed)
        url = 'https://api.daac.asf.alaska.edu/services/search/' \
            'param?granule_list='
        url += safe_dir[:-6]
        data = requests.get(url).content
        data = json.dumps(xmltodict.parse(data.decode()))
        string = json.loads(data)
        url_str = string['metalink']['files']['file'][1]['resources']['url'] \
            ['#text']

    ### Split data files into bursts
    granule_info = {}
    granule_info['platform'] = platform
    granule_info['orbitNumber'] = abolute_orbit_number
    granule_info['relativeOrbitNumber'] = relative_orbit_number
    granule_info['trackNumber'] = relative_orbit_number
    granule_info['passDirection'] = orbit_direction
    granule_info['ascendingNodeTime'] = ascending_node_time
    granule_info['granule'] = safe_dir[:-6]
    granule_info['urlOfFrame'] = url_str
    granule_info['ipf'] = ipf

    #dataframe = initializeGeoDataFrame()
    dataframe = pd.DataFrame()

    ### Extraction burst information for all IW swaths
    burst_info = \
        annotation2burst_location(annotation_iw1, burst_map, granule_info)
    dataframe = pd.concat([dataframe, burst_info], ignore_index=True)
    burst_info = \
        annotation2burst_location(annotation_iw2, burst_map, granule_info)
    dataframe = pd.concat([dataframe, burst_info], ignore_index=True)
    burst_info = \
        annotation2burst_location(annotation_iw3, burst_map, granule_info)
    dataframe = pd.concat([dataframe, burst_info], ignore_index=True)
    dataframe = fix_data_types(dataframe)
    gdf = gpd.GeoDataFrame(dataframe, crs='EPSG:4326')
    gdf.set_geometry('geometry')

    return gdf


def fix_data_types(dataframe):
    """Fix data types"""

    dataframe['insertDate'] = pd.to_datetime(dataframe['insertDate'])
    dataframe['productionDateTime'] = \
        pd.to_datetime(dataframe['productionDateTime'])
    dataframe['beginningDateTime'] = \
        pd.to_datetime(dataframe['beginningDateTime'])
    dataframe['endingDateTime'] = pd.to_datetime(dataframe['endingDateTime'])

    dataframe['startLineInSwath'] = \
        dataframe['startLineInSwath'].astype('int32')
    dataframe['endLineInSwath'] = dataframe['endLineInSwath'].astype('int32')

    dataframe['numberOfLines'] = dataframe['numberOfLines'].astype('int32')
    dataframe['numberOfSamples'] = dataframe['numberOfSamples'].astype('int32')
    dataframe['startingRange'] = dataframe['startingRange'].astype('float32')
    dataframe['sensingStart'] = pd.to_datetime(dataframe['sensingStart'])
    dataframe['sensingStop'] = pd.to_datetime(dataframe['sensingStop'])
    dataframe['burstStartUTC'] = pd.to_datetime(dataframe['burstStartUTC'])
    dataframe['burstStopUTC'] = pd.to_datetime(dataframe['burstStopUTC'])
    dataframe['trackNumber'] = dataframe['trackNumber'].astype('int32')
    dataframe['frameNumber'] = dataframe['frameNumber'].astype('int32')
    dataframe['orbitNumber'] = dataframe['orbitNumber'].astype('int32')
    dataframe['swathNumber'] = dataframe['swathNumber'].astype('int32')
    dataframe['burstNumber'] = dataframe['burstNumber'].astype('int32')
    dataframe['azimuthSteeringRate'] = \
        dataframe['azimuthSteeringRate'].astype('float32')
    dataframe['rangePixelSize'] = dataframe['rangePixelSize'].astype('float32')
    dataframe['rangeSamplingRate'] = \
        dataframe['rangeSamplingRate'].astype('float32')
    dataframe['azimuthTimeInterval'] = \
        dataframe['azimuthTimeInterval'].astype('float32')
    dataframe['radarWavelength'] = \
        dataframe['radarWavelength'].astype('float32')
    dataframe['terrainHeight'] = dataframe['terrainHeight'].astype('float32')
    dataframe['prf'] = dataframe['prf'].astype('float32')
    dataframe['firstValidLine'] = \
        dataframe['firstValidLine'].astype('int32')
    dataframe['numValidLines'] = dataframe['numValidLines'].astype('int32')
    dataframe['firstValidSample'] = \
        dataframe['firstValidSample'].astype('int32')
    dataframe['numValidSamples'] = \
        dataframe['numValidSamples'].astype('int32')
    dataframe['rangeWindowCoefficient'] = \
        dataframe['rangeWindowCoefficient'].astype('float32')
    dataframe['azimuthWindowCoefficient'] = \
        dataframe['azimuthWindowCoefficient'].astype('float32')

    dataframe['absoluteESAburstID'] = \
        dataframe['absoluteESAburstID'].astype('int32')
    dataframe['relativeESAburstID'] = \
        dataframe['relativeESAburstID'].astype('int32')
    dataframe['timeSinceAnxNode'] = \
        dataframe['timeSinceAnxNode'].astype('float32')
    dataframe['ascendingNodeTime'] = \
        pd.to_datetime(dataframe['ascendingNodeTime'])
    dataframe['hasLand'] = dataframe['hasLand'].astype('bool')
    dataframe['crossesDateLine'] = dataframe['crossesDateLine'].astype('bool')
    dataframe['landFraction'] = dataframe['landFraction'].astype('float32')

    dataframe['pointLongitude_1'] = \
        dataframe['pointLongitude_1'].astype('float32')
    dataframe['pointLatitude_1'] = \
        dataframe['pointLatitude_1'].astype('float32')
    dataframe['pointLongitude_2'] = \
        dataframe['pointLongitude_2'].astype('float32')
    dataframe['pointLatitude_2'] = \
        dataframe['pointLatitude_2'].astype('float32')
    dataframe['pointLongitude_3'] = \
        dataframe['pointLongitude_3'].astype('float32')
    dataframe['pointLatitude_3'] = \
        dataframe['pointLatitude_3'].astype('float32')
    dataframe['pointLongitude_4'] = \
        dataframe['pointLongitude_4'].astype('float32')
    dataframe['pointLatitude_4'] = \
        dataframe['pointLatitude_4'].astype('float32')
    dataframe['pointLongitude_5'] = \
        dataframe['pointLongitude_5'].astype('float32')
    dataframe['pointLatitude_5'] = \
        dataframe['pointLatitude_5'].astype('float32')

    return dataframe


def parse_line(line, output_type):
    """Parse a line"""

    return_value = None
    if line.count(' - ') == 2:
        (datestamp, log_level, remainder) = line.split(' - ')
        if output_type == 'dateString':
            return_value = datetime.strptime(datestamp,
                '%m/%d/%Y %H:%M:%S %p').isoformat() + 'Z'
        elif output_type == 'log_level':
            return_value = log_level.split()
        if ':' in remainder:
            (key, value) = remainder.split(':', 1)
            if output_type == 'key':
                return_value = key.strip()
            elif output_type == 'value':
                return_value = value.strip()
    elif line.count(' - ') == 0 and line.count(': ') == 1:
        (key, value) = line.split(': ')
        if output_type == 'key':
            return_value = key
        elif output_type == 'value':
            return_value = value.strip()

    return return_value


def get_sentinel_files(safe, meta):
    """Get Sentinel file information"""

    files = []
    doc = et.fromstring(meta['dataObjectSection'])
    data_objects = doc.xpath('/dataObjectSection/dataObject')
    for i in range(len(data_objects)):
        file = {}
        href = doc.xpath('/dataObjectSection/dataObject[{0}]/byteStream/' \
            'fileLocation/@href'.format(i+1))[0]
        file_parts = os.path.basename(href).split('-')
        file['name'] = os.path.join(safe, href[2:])
        file['index'] = doc.xpath('/dataObjectSection/dataObject[{0}]/' \
            '@ID'.format(i+1))[0]
        if file['index'] == 'mapoverlay':
            file['type'] = 'preview'
        elif file['index'] == 'productpreview':
            file['type'] = 'preview'
        elif file['index'] == 'quicklook':
            file['type'] = 'preview'
        elif file['index'].startswith('product'):
            file['type'] = 'annotation'
            file['swath'] = file_parts[1].upper()
            file['polarization'] = file_parts[3].upper()
            file['start_time'] = file_parts[4]
            file['stop_time'] = file_parts[5]
        elif file['index'].startswith('calibration'):
            file['type'] = 'calibration'
            file['swath'] = file_parts[2].upper()
            file['polarization'] = file_parts[4].upper()
            file['start_time'] = file_parts[5]
            file['stop_time'] = file_parts[6]
        elif file['index'].startswith('noise'):
            file['type'] = 'noise'
            file['swath'] = file_parts[2].upper()
            file['polarization'] = file_parts[4].upper()
            file['start_time'] = file_parts[5]
            file['stop_time'] = file_parts[6]
        elif file['index'].startswith('rfi'):
            file['type'] = 'rfi'
            file['swath'] = file_parts[2].upper()
            file['polarization'] = file_parts[4].upper()
            file['start_time'] = file_parts[5]
            file['stop_time'] = file_parts[6]
        else:
            file['type'] = 'measurement'
            file['swath'] = file_parts[1].upper()
            file['polarization'] = file_parts[3].upper()
            file['start_time'] = file_parts[4]
            file['stop_time'] = file_parts[5]
        file['size'] = int(doc.xpath('/dataObjectSection/' \
            'dataObject[{0}]/byteStream/@size'.format(i+1))[0])
        file['checksum'] = doc.xpath('/dataObjectSection/' \
            'dataObject[{0}]/byteStream/checksum'.format(i+1))[0].text
        files.append(file)

    return files



#!/usr/bin/python3

"""Import packages"""
import json
from datetime import datetime
import collections
import re
import os
import glob
import lxml.etree as et
from lxml import objectify
import pandas as pd
import numpy as np
from osgeo import gdal, osr

from .util import get_value, rreplace, unique_list, \
  get_params_dataframe, params2dict_list, upgrade_dictionary2level, \
  merge_dictionary_params, get_level_params_list
from .sentinel_metadata import parse_line, get_latlon_extent, \
  read_manifest_file, read_annotation_file


def meta_project2epsg(meta_dict):
    """Determine EPSG code from projection type"""

    projection = meta_dict['projection']['type']
    if projection == 'UNIVERSAL_TRANSVERSE_MERCATOR':
        hemisphere = meta_dict['projection']['hem']
        zone = int(meta_dict['projection']['param']['utm']['zone'])
        if hemisphere == 'N':
            epsg = 32600 + zone
        else:
            epsg = 32700 + zone
    return epsg


def update_template(template, list_params):
    """Update template"""

    count = len(template)
    for list_param in list_params:
        indices = [i for i, x in enumerate(template) if x == list_param]
        indices.append(count)
        for i in range(len(indices)-1):
            for k in range(indices[i], indices[i+1]):
                if list_param in template[k]:
                    new_element = f'{list_param}[{i+1}]'
                    template[k] = \
                        rreplace(template[k], list_param, new_element)

    return template


def iso_template2lists(excel_file, worksheet):
    """Convert ISO template to lists"""

    ### ISO metadata structure
    meta = pd.read_excel(excel_file, sheet_name=worksheet)
    (row_count, level_count) = meta.shape
    meta_type = list(meta['Type'])
    del meta['Type']
    meta_value = meta['Value']
    del meta['Value']
    if worksheet == 'ISO Metadata Structure':
        iso_attributes = list(meta['Attribute'])
        del meta['Attribute']
        iso_attribute_values = list(meta['AttributeValue'])
        del meta['AttributeValue']
    else:
        iso_attributes = []
        for i in range(row_count):
            iso_attributes.append('nan')
    (row_count, level_count) = meta.shape

    ### Read all entries in the template
    iso_template = []
    for i in range(row_count):
        param = ''
        for k in range(level_count):
            element = str(meta.at[i, k])
            if element != 'nan':
                param += element
                if k < (level_count-1) and str(meta.at[i, k+1]) != 'nan':
                    param += '/'
        iso_template.append(param)

    ### Collect elements that contain lists
    duplicates = []
    for i in range(level_count):
        for k in range(row_count):
            occurences = [x.start() for x in re.finditer('/', iso_template[k])]
            if len(occurences) == i:
                if 'list' in meta_type[k]:
                    duplicates.append(iso_template[k])
        list_params = [item for item, \
            count in collections.Counter(duplicates).items() if count > 0]
        if len(list_params) > 0:
            iso_template = update_template(iso_template, list_params)

    ### Read parameters out of template
    iso_params = []
    iso_values = []
    for i in range(row_count):
        if 'attribute' in meta_type[i]:
            attrib = iso_template[i] + '/' + str(iso_attributes[i])
            iso_params.append(attrib)
            iso_values.append(iso_attribute_values[i])
        if 'value' in meta_type[i]:
            iso_params.append(iso_template[i])
            iso_values.append(meta_value[i])

    return (iso_template, iso_params, iso_values)


def merge_iso_dem_template(iso_template, dem_template):
    """Merge ISO and DEM templates"""

    ### Analyze ISO template
    max_identification_info = 0
    max_identification_index = 0
    max_content_info = 0
    max_content_index = 0
    for i in range(len(iso_template)):
        test_level = iso_template[i][-2:-1]
        if iso_template[i][:-3].endswith('identificationInfo'):
            if max_identification_info < int(test_level):
                max_identification_info = int(test_level)
        if 'identificationInfo' in iso_template[i] and \
            max_identification_index < i:
            max_identification_index = i
        if iso_template[i][:-3].endswith('contentInfo'):
            if max_content_info < int(test_level):
                max_content_info = int(test_level)
        if 'contentInfo' in iso_template[i] and max_content_index < i:
            max_content_index = i

    ### Analyze DEM template
    max_index = 0
    for i in range(len(dem_template)):
        if 'identificationInfo' in dem_template[i] and max_index < i:
            max_index = i

    ### Combine ISO and DEM templates
    merged_template = []
    for i in range(max_identification_index+1):
        merged_template.append(iso_template[i])
    original_string = 'identificationInfo[1]'
    replace_string = \
        ('identificationInfo[{0}]'.format(max_identification_info+1))
    for i in range(max_index+1):
        if 'identificationInfo' in dem_template[i]:
            dem_info = dem_template[i].replace(original_string, replace_string)
            merged_template.append(dem_info)
    for i in range(max_identification_index+1, max_content_index+1):
        merged_template.append(iso_template[i])
    original_string = 'contentInfo[1]'
    replace_string = ('contentInfo[{0}]'.format(max_identification_info+1))
    for i in range(max_index+1, len(dem_template)):
        dem_info = dem_template[i].replace(original_string, replace_string)
        merged_template.append(dem_info)
    for i in range(max_content_index+1, len(iso_template)):
        merged_template.append(iso_template[i])

    return merged_template


def add_dem_lists(iso_template, iso_params, iso_values, dem_template,
    dem_params, dem_values):
    """Add DEM lists"""

    ### Analyze ISO template
    max_identification_info = 0
    max_identification_index = 0
    max_content_info = 0
    max_content_index = 0
    for i in range(len(iso_template)):
        test_level = iso_template[i][-2:-1]
        if iso_template[i][:-3].endswith('identificationInfo'):
            if max_identification_info < int(test_level):
                max_identification_info = int(test_level)
        if 'identificationInfo' in iso_template[i] and \
            max_identification_index < i:
            max_identification_index = i
        if iso_template[i][:-3].endswith('contentInfo'):
            if max_content_info < int(test_level):
                max_content_info = int(test_level)
        if 'contentInfo' in iso_template[i] and max_content_index < i:
            max_content_index = i

    ### Analyze DEM template
    max_index = 0
    for i in range(len(dem_template)):
        if 'identificationInfo' in dem_template[i] and max_index < i:
            max_index = i

    ### Combine ISO and DEM templates
    merged_template = []
    for i in range(max_identification_index+1):
        merged_template.append(iso_template[i])
    original_string = 'identificationInfo[1]'
    replace_string = \
        ('identificationInfo[{0}]'.format(max_identification_info+1))
    for i in range(max_index+1):
        if 'identificationInfo' in dem_template[i]:
            dem_info = dem_template[i].replace(original_string, replace_string)
            merged_template.append(dem_info)
    for i in range(max_identification_index+1, max_content_index+1):
        merged_template.append(iso_template[i])
    original_string = 'contentInfo[1]'
    replace_string = \
        ('contentInfo[{0}]'.format(max_identification_info+1))
    for i in range(max_index+1, len(dem_template)):
        dem_info = dem_template[i].replace(original_string, replace_string)
        merged_template.append(dem_info)
    for i in range(max_content_index+1, len(iso_template)):
        merged_template.append(iso_template[i])

    ### Analyze ISO params and values
    max_identification_index = 0
    max_content_index = 0
    for i in range(len(iso_params)):
        if 'identificationInfo' in iso_params[i] and \
            max_identification_index < i:
            max_identification_index = i
        if 'contentInfo' in iso_params[i] and max_content_index < i:
            max_content_index = i

    ### Analyze DEM params
    max_index = 0
    for i in range(len(dem_params)):
        if 'identificationInfo' in dem_params[i] and max_index < i:
            max_index = i

    ### Combine ISO and DEM params and values
    merged_params = []
    merged_values = []
    for i in range(max_identification_index+1):
        merged_params.append(iso_params[i])
        merged_values.append(iso_values[i])
    original_string = 'identificationInfo[1]'
    replace_string = \
        ('identificationInfo[{0}]'.format(max_identification_info+1))
    for i in range(max_index+1):
        dem_info = dem_params[i].replace(original_string, replace_string)
        merged_params.append(dem_info)
        merged_values.append(dem_values[i])
    for i in range(max_identification_index+1, max_content_index+1):
        merged_params.append(iso_params[i])
        merged_values.append(iso_values[i])
    original_string = 'contentInfo[1]'
    replace_string = ('contentInfo[{0}]'.format(max_identification_info+1))
    for i in range(max_index+1, len(dem_params)):
        dem_info = dem_params[i].replace(original_string, replace_string)
        merged_params.append(dem_info)
        merged_values.append(dem_values[i])
    for i in range(max_content_index+1, len(iso_params)):
        merged_params.append(iso_params[i])
        merged_values.append(iso_values[i])

    return (merged_template, merged_params, merged_values)


def iso_attributes2lists(excel_file):
    """Convert ISO attributes to lists"""

    ### ISO metadata structure
    meta = pd.read_excel(excel_file, sheet_name='ISO Metadata Structure')
    iso_attributes = list(meta['Attribute'])
    iso_attribute_values = list(meta['AttributeValue'])
    (iso_template, _, _) = \
        iso_template2lists(excel_file, 'ISO Metadata Structure')

    return (iso_attributes, iso_attribute_values, iso_template)


def iso_xml_structure(excel_file, iso_template, iso_params, iso_values,
    ns_flag):
    """Build ISO XML structure"""

    ### Get name_space look up table
    ns_lut = pd.read_excel(excel_file, sheet_name='Namespaces LUT')
    ns_pairs = dict(zip(ns_lut['Parameter'], ns_lut['Namespace']))

    ### Build name_space
    ns_lut = pd.read_excel(excel_file, sheet_name='Namespaces URL')
    ns_lut = ns_lut.set_index('Namespace')
    ns_xsi = {'xsi': ns_lut.at['xsi','URL']}
    ns_gmd = {'gmd': ns_lut.at['gmd','URL']}
    ns_gco = {'gco': ns_lut.at['gco','URL']}
    ns_xs = {'xs': ns_lut.at['xs','URL']}
    ns_eos = {'eos': ns_lut.at['eos','URL']}
    ns_echo = {'echo': ns_lut.at['echo','URL']}
    ns_xlink = {'xlink': ns_lut.at['xlink','URL']}
    ns_gml = {'gml': ns_lut.at['gml','URL']}
    ns_gmi = {'gmi': ns_lut.at['gmi','URL']}
    ns_gmx = {'gmx': ns_lut.at['gmx','URL']}
    ns_all = dict(
        list(ns_xsi.items()) +
        list(ns_gmd.items()) +
        list(ns_gco.items()) +
        list(ns_xs.items()) +
        list(ns_eos.items()) +
        list(ns_echo.items()) +
        list(ns_xlink.items()) +
        list(ns_gml.items()) +
        list(ns_gmi.items()) +
        list(ns_gmx.items())
    )

    iso_elements = []
    iso_element_count = []
    row_count = len(iso_template)
    level_count = 0
    for i in range(row_count):
        meta_elements = iso_template[i].split('/')
        if len(meta_elements) > level_count:
            level_count = len(meta_elements)
        iso_elements.append(meta_elements)
        iso_element_count.append(len(meta_elements))
    level_elements = []
    for k in range(level_count):
        meta = []
        for i in range(row_count):
            if k < len(iso_elements[i]):
                meta.append(iso_elements[i][k])
        level_elements.append(unique_list(meta))

    ### Build XMl element tree from nested dictionary
    name_space = ('{%s}' % ns_lut.at[ns_pairs['DS_Series'],'URL'])
    if ns_flag:
        root = \
            et.Element('{0}{1}'.format(name_space, 'DS_Series'), nsmap=ns_all)
    else:
        root = et.Element('DS_Series')

    ## Build structure
    ref = {}
    ref['DS_Series'] = root
    for i in range(1, level_count):
        indices = [k for k, x in enumerate(iso_element_count) if x == i+1]
        for k in range(len(indices)):
            element = iso_template[indices[k]]
            parent = ref[element.rsplit('/',1)[0]]
            param = element.rsplit('/',1)[1]
            if 'EOS_AdditionalAttributes' in element:
                name_space = ('{%s}' % ns_lut.at['eos','URL'])
            elif ('acquisitionInformation' in element) and \
                param in ('identifier','description','type'):
                name_space = ('{%s}' % ns_lut.at['gmi','URL'])
            else:
                name_space = ('{%s}' % \
                    ns_lut.at[ns_pairs[param.split('[')[0]],'URL'])
            if ns_flag:
                child = et.SubElement(parent, '{0}{1}'.format(name_space,
                    param.split('[')[0], ns_map=ns_all))
            else:
                child = et.SubElement(parent, param.split('[')[0])
            ref[element] = child

    ## Add values
    for i in range(len(iso_params)):
        element = iso_params[i]
        param = element.rsplit('/',1)[1]
        if (param == 'value') and (('result/value' not in element) or \
            ('EOS_AdditionalAttribute/value' not in element)):
            continue
        elif param == 'id':
            reference = ref[element.rsplit('/',1)[0]]
            name_space = \
                ('{%s}' % ns_lut.at[ns_pairs[param.split('[')[0]],'URL'])
            if ns_flag:
                ns_param = '{0}{1}'.format(name_space, param.split('[')[0],
                    ns_map=ns_all)
            else:
                ns_param = param.split('[')[0]
            reference.attrib[ns_param] = str(iso_values[i])
        elif param in ('href','nilReason'):
            reference = ref[element.rsplit('/',1)[0]]
            name_space = \
                ('{%s}' % ns_lut.at[ns_pairs[param.split('[')[0]],'URL'])
            if ns_flag:
                ns_param = '{0}{1}'.format(name_space, param.split('[')[0],
                    ns_map=ns_all)
            else:
                ns_param = param.split('[')[0]
            reference.attrib[ns_param] = str(iso_values[i])
        elif param == 'type':
            reference = ref[element.rsplit('/',1)[0]]
            name_space = ('{%s}' % ns_lut.at['xsi','URL'])
            if ns_flag:
                ns_param = '{0}{1}'.format(name_space, param.split('[')[0],
                    ns_map=ns_all)
            else:
                ns_param = param.split('[')[0]
            reference.attrib[ns_param] = str(iso_values[i])
        elif param == 'uom':
            reference = ref[element.rsplit('/',1)[0]]
            reference.attrib[param] = str(iso_values[i])
            reference.text = str(iso_values[i-1])
        elif param.split('[')[0] in ns_pairs:
            name_space = \
                ('{%s}' % ns_lut.at[ns_pairs[param.split('[')[0]],'URL'])
            reference = ref[element]
            reference.text = str(iso_values[i])
            if 'Code' in param:
                if 'EOS' in element:
                    code_url = 'http://earthdata.nasa.gov/metadata/' \
                        'resources/codelist.xml#' + param
                else:
                    code_url = 'https://cdn.earthdata.nasa.gov/iso/' \
                        'resources/Codelist/gmxCodelists.xml#' + param
                reference.attrib['codeList'] = code_url
                reference.attrib['codeListValue'] = str(iso_values[i])
            if param == 'RecordType':
                name_space = ('{%s}' % ns_lut.at['xlink','URL'])
                if ns_flag:
                    ns_param = '{0}{1}'.format(name_space, 'href',
                        ns_map=ns_all)
                else:
                    ns_param = param.split('[')[0]
                reference.attrib[ns_param] = "http://earthdata.nasa.gov/" \
                    "schemas/eos/eos.xsd#xpointer(//" \
                    "element[@name='EOS_AdditionalAttributes'])"

    return root


def add_dem_attributes(iso_attributes, dem_attributes, iso_template,
    dem_template):
    """Add DEM attributes"""

    ### Analyze ISO template
    max_identification_info = 0
    max_identification_index = 0
    max_content_info = 0
    max_content_index = 0
    for i in range(len(iso_template)):
        test_level = iso_template[i][-2:-1]
        if iso_template[i][:-3].endswith('identificationInfo'):
            if max_identification_info < int(test_level):
                max_identification_info = int(test_level)
        if 'identificationInfo' in iso_template[i] and \
            max_identification_index < i:
            max_identification_index = i
        if iso_template[i][:-3].endswith('contentInfo'):
            if max_content_info < int(test_level):
                max_content_info = int(test_level)
        if 'contentInfo' in iso_template[i] and max_content_index < i:
            max_content_index = i

    ### Analyze DEM template
    max_index = 0
    for i in range(len(dem_template)):
        if 'identificationInfo' in dem_template[i] and max_index < i:
            max_index = i

    ### Combine ISO and DEM attributes
    merged_attributes = []
    for i in range(max_identification_index+1):
        merged_attributes.append(iso_attributes[i])
    for i in range(max_index+1):
        merged_attributes.append(dem_attributes[i])
    for i in range(max_identification_index+1, max_content_index+1):
        merged_attributes.append(iso_attributes[i])
    for i in range(max_index+1, len(dem_attributes)):
        merged_attributes.append(dem_attributes[i])
    for i in range(max_content_index+1, len(iso_attributes)):
        merged_attributes.append(iso_attributes[i])

    return merged_attributes


def iso_dictionary_structure(excel_file, dem_file, meta_template, meta_params,
  meta_values):
    """Build ISO dictionary structure"""

    ### Read attributes and their values from Excel spreadsheet
    (iso_attributes, _, iso_template) = iso_attributes2lists(excel_file)
    if dem_file:
        (dem_attributes, _, dem_template) = iso_attributes2lists(dem_file)
        iso_attributes = add_dem_attributes(iso_attributes, dem_attributes[5:],
            iso_template, dem_template[5:])

    ### Build dataframe with parameters
    df_params = get_params_dataframe(meta_params)
    (_, old_level) = df_params.shape
    old_level -= 1
    dict_list = \
        params2dict_list(df_params, meta_params, meta_values, old_level)

    ### Build dictionary structure - one level at a time
    for new_level in reversed(range(old_level)):

        new_dict_list = \
            params2dict_list(df_params, meta_params, meta_values, new_level)
        dict_list += new_dict_list
        dict_list = upgrade_dictionary2level(dict_list, old_level, new_level)
        level_params = \
            get_level_params_list(meta_template, iso_attributes, new_level)
        dict_list = merge_dictionary_params(dict_list, level_params, old_level,
            new_level)
        old_level = new_level

    iso_dict = {}
    iso_dict[dict_list[0]['key']] = dict_list[0]['dict']

    return iso_dict


def umm_dictionary_structure(meta_template, meta_params, meta_values):
    """Build UMM dictionary structure"""

    ### Build dataframe with parameters
    df_params = get_params_dataframe(meta_params)
    (_, old_level) = df_params.shape
    old_level -= 1
    dict_list = \
        params2dict_list(df_params, meta_params, meta_values, old_level)

    ### Build dictionary structure - one level at a time
    for new_level in reversed(range(old_level)):

        new_dict_list = \
            params2dict_list(df_params, meta_params, meta_values, new_level)
        dict_list += new_dict_list
        dict_list = upgrade_dictionary2level(dict_list, old_level, new_level)
        level_params = get_level_params_list(meta_template, None, new_level)
        dict_list = merge_dictionary_params(dict_list, level_params, old_level,
            new_level)
        old_level = new_level

    iso_dict = {}
    iso_dict[dict_list[0]['key']] = dict_list[0]['dict']

    return iso_dict


def meta_json_file(meta_structure, json_file):
    """Write metadata structure to JSON file"""

    ### Write JSON file
    with open(json_file, 'w') as file:
        json.dump(meta_structure, file, indent=2)


def meta_xml_file(meta_structure, xml_file):
    """Write metaadata structure to XML File"""

    ### Write XML file
    with open(xml_file, 'wb') as file:
        file.write(et.tostring(meta_structure,
            xml_declaration=True, encoding='utf-8',
            pretty_print=True))


def generate_product_dictionary(product_type, data_type, meta_file,
    log_file):
    """Generate product dictionary"""

    if product_type == 'gamma_rtc':
        meta = gamma_rtc_log2meta(data_type, meta_file, log_file)

    return meta


def get_metadata_values(meta_file, params):
    """Get metadata values"""

    product_params = []

    parser = et.XMLParser(remove_blank_text=True)
    tree = et.parse(meta_file, parser)
    root = tree.getroot()
    for elem in root.getiterator():
        if len(elem.attrib) > 0:
            key = elem.attrib.keys()
            item = elem.attrib.items()
            (key, value) = item[0]
            if '}' in key:
                del elem.attrib[key]
                key = key.split('}')[1]
                elem.set(key, value)
        if not hasattr(elem.tag, 'find'):
            continue
        i = elem.tag.find('}')
        if i >= 0:
            elem.tag = elem.tag[i+1:]
    objectify.deannotate(root, cleanup_name_spaces=True)

    for i in range(len(params)):
        try:
            param = tree.xpath('/'+params[i])[0].text
            product_params.append(get_value(param))
        except:
            (element, attribute) = params[i].rsplit('/',1)
            if attribute != 'value':
                element_attribute = ('/'+element+'/@'+attribute)
                param = tree.xpath(element_attribute)[0]
                product_params.append(get_value(param))
            else:
                param = tree.xpath('/'+element)[0].text
                product_params.append(get_value(param))

    return product_params


def product_dictionary2values(values, prod_dict):
    """Convert product dictionary to values"""

    dict_keys = list(prod_dict.keys())
    prod_values = []
    for i in range(len(values)):
        keys = []
        for k in range(len(dict_keys)):
            if str(values[i]).count(dict_keys[k]) > 0:
                keys.append(dict_keys[k])
        for key in keys:
            start = values[i].index('{# ')
            stop = values[i].index(' #}')
            parameter = values[i][start:stop+3]
            values[i] = \
                values[i].replace(parameter, str(prod_dict[key]))
            try:
                if isinstance(prod_dict[key], float):
                    values[i] = float(values[i])
                elif isinstance(prod_dict[key], bool):
                    values[i] = bool(values[i])
                elif isinstance(prod_dict[key], int):
                    values[i] = int(values[i])
            except:
                pass

        prod_values.append(values[i])

    return prod_values


def granule_iso2umm_values(iso_params, iso_values, iso2umm, umm_values):
    """Convert granule ISO to UMM values"""

    (param_count, _) = iso2umm.shape
    for k in range(len(umm_values)):
        for i in range(param_count):
            umm_param_string = ('{# UMM_%s #}' % iso2umm.at[i, 'UMM'])
            if umm_param_string == umm_values[k]:
                iso_index = iso_params.index(iso2umm.at[i, 'ISO'])
                umm_values[k] = iso_values[iso_index]

    return umm_values


def properties2values(template, properties):
    """Convert properties to values"""

    values = []
    for param in template:
        if param in properties.keys():
            try:
                if isinstance(properties[param], float):
                    values.append(float(properties[param]))
                elif isinstance(properties[param], bool):
                    values.append(bool(properties[param]))
                elif isinstance(properties[param], int):
                    values.append(int(properties[param]))
                else:
                    values.append(properties[param])
            except:
                pass
        else:
            values.append(param)

    return values


def clean_json_structure(iso_structure):
    """Clean JSON structure"""

    meta = iso_structure['DS_Series']['composedOf']['DS_DataSet']['has'] \
        ['MI_Metadata']

    identification_info = meta['identificationInfo']
    new_identification = []
    for i in range(len(identification_info)):
        value = identification_info[i]['MD_DataIdentification']['citation'] \
            ['CI_Citation']['title']['FileName']
        if not value.startswith('{# ISO'):
            new_identification.append(identification_info[i])
    meta['identificationInfo'] = new_identification

    content_info = meta['contentInfo']
    new_content = []
    for i in range(len(content_info)):
        mi_band = \
            content_info[i]['MD_CoverageDescription']['dimension']['MI_Band']
        if 'maxValue' in mi_band:
            value = mi_band['maxValue']['Real']
        elif 'otherPropertyType' in mi_band:
            value = mi_band['otherProperty']['Record']['additionalAttributes'] \
                ['EOS_AdditionalAttributes']['additionalAttribute'][0] \
                ['EOS_AdditionalAttribute']['value']['CharacterString']
        if isinstance(value, float):
            new_content.append(content_info[i])
    meta['contentInfo'] = new_content

    return iso_structure


def clean_xml_structure(iso_file):
    """Clean XML structure"""

    ns_gmd = {'gmd' : 'http://www.isotc211.org/2005/gmd'}
    ns_gco = {'gco' : 'http://www.isotc211.org/2005/gco'}
    ns_eos = {'eos' : 'http://earthdata.nasa.gov/schema/eos'}
    ns_gmi = {'gmi' : 'http://www.isotc211.org/2005/gmi'}
    ns_gmx = {'gmx' : 'http://www.isotc211.org/2005/gmx'}
    ns_all = dict(
        list(ns_gmd.items()) +
        list(ns_gco.items()) +
        list(ns_eos.items()) +
        list(ns_gmi.items()) +
        list(ns_gmx.items())
    )

    parser = et.XMLParser(remove_blank_text=True)
    doc = et.parse(iso_file, parser)
    meta = doc.xpath('/gmd:DS_Series/gmd:composedOf/gmd:DS_DataSet' \
        '/gmd:has/gmi:MI_Metadata', namespaces=ns_all)[0]
    identification_info_count = 0
    content_info_count = 0
    remove = []
    for i in range(len(meta)):
        if 'identificationInfo' in meta[i].tag:
            file_name = doc.xpath('/gmd:DS_Series/gmd:composedOf/' \
                'gmd:DS_DataSet/gmd:has/gmi:MI_Metadata/' \
                'gmd:identificationInfo[{0}]/gmd:MD_DataIdentification/' \
                'gmd:citation/gmd:CI_Citation/gmd:title/gmx:FileName' \
                .format(identification_info_count+1), namespaces=ns_all)[0].text
            if file_name.startswith('{# ISO'):
                remove.append(i)
            identification_info_count += 1
        elif 'contentInfo' in meta[i].tag:
            mi_band = doc.xpath('/gmd:DS_Series/gmd:composedOf/' \
                'gmd:DS_DataSet/gmd:has/gmi:MI_Metadata/' \
                'gmd:contentInfo[{0}]/gmd:MD_CoverageDescription/' \
                'gmd:dimension/gmi:MI_Band'.format(content_info_count+1),
                namespaces=ns_all)[0]
            if 'maxValue' in mi_band[0].tag:
                value = doc.xpath('/gmd:DS_Series/gmd:composedOf/' \
                    'gmd:DS_DataSet/gmd:has/gmi:MI_Metadata/' \
                    'gmd:contentInfo[{0}]/gmd:MD_CoverageDescription/' \
                    'gmd:dimension/gmi:MI_Band/gmd:maxValue/gco:Real' \
                    .format(content_info_count+1), namespaces=ns_all)[0].text
            elif 'otherPropertyType' in mi_band[0].tag:
                value = doc.xpath('/gmd:DS_Series/gmd:composedOf/' \
                    'gmd:DS_DataSet/gmd:has/gmi:MI_Metadata/' \
                    'gmd:contentInfo[{0}]/gmd:MD_CoverageDescription/' \
                    'gmd:dimension/gmi:MI_Band/eos:otherProperty/gco:Record/' \
                    'eos:additionalAttributes/eos:EOS_AdditionalAttributes/' \
                    'eos:additionalAttribute[1]/eos:EOS_AdditionalAttribute/' \
                    'eos:value/eos:CharacterString' \
                    .format(content_info_count+1), namespaces=ns_all)[0].text
            try:
                float(value)
            except ValueError:
                remove.append(i)
            content_info_count += 1

    for i in range(len(remove)-1,-1,-1):
        del meta[remove[i]]

    return doc


def gamma_rtc_log2meta(data_type, meta_file, log_file):
    """Convert GAMMA RTC log to metadata structure"""

    ns_safe = {'safe': 'http://www.esa.int/safe/sentinel-1.0'}
    ns_all = dict(
        list(ns_safe.items())
    )

    ### Read information from log file
    with open(log_file, encoding='utf-8') as file:
        lines = file.readlines()

    meta = {}
    meta['ISO_RTC_metadataCreationTime'] = datetime.utcnow().isoformat() + 'Z'
    meta['ISO_RTC_productCreationTime'] = parse_line(lines[0], 'dateString')
    meta['ISO_RTC_radiometry'] = parse_line(lines[3], 'value')
    meta['ISO_RTC_scale'] = parse_line(lines[4], 'value')
    meta['ISO_RTC_DEM_type'] = parse_line(lines[13], 'value').upper()
    for line in lines:
        value = parse_line(line, 'value')
        if value and value.startswith('geoid file'):
            meta['ISO_RTC_GammaVersion'] = \
                [i for i in value.split('/') \
                    if 'GAMMA_SOFTWARE' in i][0].split('-')[1]
        elif value and value.startswith('number of azimuth looks'):
            meta['ISO_RTC_azimuthLooks'] = int(value.split(':')[1])
        elif value and value.startswith('number of range looks'):
            meta['ISO_RTC_rangeLooks'] = int(value.split(':')[1])
    meta['ISO_RTC_XML_filename'] = log_file.replace('.log','.iso.xml')

    if data_type == 'sentinel':

        ### Extract more metadata from manifest file
        input_path = os.path.dirname(meta_file)
        meta['ISO_RTC_inputGranule'] = os.path.basename(input_path)[:-5]
        manifest_file = meta_file
        manifest = read_manifest_file(manifest_file)
        doc = et.fromstring(manifest['metadataSection'])
        meta['ISO_RTC_startTime'] = \
            doc.xpath('/metadataSection/metadataObject' \
            '[@ID="acquisitionPeriod"]/metadataWrap/xmlData/' \
            'safe:acquisitionPeriod/safe:startTime',
            namespaces=ns_all)[0].text + 'Z'
        meta['ISO_RTC_stopTime'] = \
            doc.xpath('/metadataSection/metadataObject' \
            '[@ID="acquisitionPeriod"]/metadataWrap/xmlData/' \
            'safe:acquisitionPeriod/safe:stopTime',
            namespaces=ns_all)[0].text + 'Z'

        ### Extract more metadata from annotation file
        annotation_path = os.path.join(input_path, 'annotation', '*.xml')
        annotation_file = glob.glob(f'{annotation_path}')[0]
        annotation = read_annotation_file(annotation_file)
        doc = et.fromstring(annotation['adsHeader'])
        meta['ISO_RTC_mission'] = \
            doc.xpath('/adsHeader/missionId')[0].text.replace('S','SENTINEL-')
        meta['ISO_RTC_beamMode'] = doc.xpath('/adsHeader/mode')[0].text
        meta['ISO_RTC_absoluteOrbitNumber'] = \
            int(doc.xpath('/adsHeader/absoluteOrbitNumber')[0].text)
        doc = et.fromstring(annotation['generalAnnotation'])
        meta['ISO_RTC_orbitPassDirection'] = \
            doc.xpath('/generalAnnotation/productInformation/pass')[0].text.lower()
        speed_of_light = 299792458.0
        frequency = float(doc.xpath('/generalAnnotation/productInformation/' \
            'radarFrequency')[0].text)
        meta['ISO_RTC_wavelength'] = speed_of_light / frequency

        ### Examine product directory for various files and calculate stats if exists
        polarization = meta['ISO_RTC_inputGranule'][14:16]
        if polarization in ('SH','DH'):
            main_pol = 'HH'
            pol_string = 'horizontal'
        elif polarization in ('SV','DV'):
            main_pol = 'VV'
            pol_string = 'vertical'

    ### Extract Hyp3 version out of README file
    readme_file = log_file.replace('.log','.README.md.txt')
    with open(readme_file, encoding='utf-8') as file:
        lines = file.readlines()
    for line in lines:
        if parse_line(line, 'key') == 'Metadata version':
            meta['ISO_RTC_productVersion'] = parse_line(line, 'value')

    ### Setting additional metadata
    meta['ISO_RTC_lookDirection'] = 'RIGHT'
    if meta['ISO_RTC_mission'] == 'ALOS':
        meta['ISO_RTC_sensor'] = 'PALSAR'
    else:
        meta['ISO_RTC_sensor'] = 'SAR'

    ## Look at product file
    product_file = log_file.replace('.log',f'_{main_pol}.tif')
    if os.path.isfile(product_file):
        meta['ISO_RTC_terrainCorrectedImage'] = os.path.basename(product_file)
        raster = gdal.Open(product_file)
        meta['ISO_RTC_azimuthCount'] = raster.RasterXSize
        band = raster.GetRasterBand(1)
        stats = band.GetStatistics(False, True)
        proj = osr.SpatialReference()
        proj.ImportFromWkt(raster.GetProjectionRef())
        if proj.GetAttrValue('AUTHORITY', 0) == 'EPSG':
            epsg = int(proj.GetAttrValue('AUTHORITY', 1))
            geo_transform = raster.GetGeoTransform()
            meta['ISO_RTC_azimuthPixelSize'] = geo_transform[1]
            meta['ISO_RTC_rangePixelSize'] = -geo_transform[5]
            meta['ISO_RTC_azimuthCount'] = raster.RasterXSize
            meta['ISO_RTC_rangeCount'] = raster.RasterYSize
        else:
            epsg = 0
        meta['ISO_RTC_minValue'] = stats[0]
        meta['ISO_RTC_maxValue'] = stats[1]
        meta['ISO_RTC_meanValue'] = stats[2]
        meta['ISO_RTC_standardDeviation'] = stats[3]
        meta['ISO_RTC_transmittedPolarization'] = pol_string
        meta['ISO_RTC_receivedPolarization'] = pol_string
        meta['ISO_RTC_epsgCode'] = epsg

    ## Look at DEM file
    dem_file = log_file.replace('.log','_dem.tif')
    if os.path.isfile(dem_file):
        meta['ISO_RTC_digitalElevationModel'] = os.path.basename(dem_file)
        raster = gdal.Open(product_file)
        band = raster.GetRasterBand(1)
        stats = band.GetStatistics(False, True)
        meta['ISO_RTC_DEM_minValue'] = stats[0]
        meta['ISO_RTC_DEM_maxValue'] = stats[1]
        meta['ISO_RTC_DEM_meanValue'] = stats[2]
        meta['ISO_RTC_DEM_standardDeviation'] = stats[3]

    ## Look at the incidence angle map
    incidence_angle_file = log_file.replace('.log','_inc.tif')
    if os.path.isfile(incidence_angle_file):
        meta['ISO_RTC_incidenceAngleMap'] = \
            os.path.basename(incidence_angle_file)
        raster = gdal.Open(product_file)
        band = raster.GetRasterBand(1)
        stats = band.GetStatistics(False, True)
        meta['ISO_RTC_incidenceMinValue'] = stats[0]
        meta['ISO_RTC_incidenceMaxValue'] = stats[1]
        meta['ISO_RTC_incidenceMeanValue'] = stats[2]
        meta['ISO_RTC_incidenceStandardDeviation'] = stats[3]

    ## Look at the layover shadow mask
    layover_shadow_mask_file = log_file.replace('.log','_ls_map.tif')
    if os.path.isfile(layover_shadow_mask_file):
        meta['ISO_RTC_layoverShadowMask'] = \
            os.path.basename(layover_shadow_mask_file)
        raster = gdal.Open(layover_shadow_mask_file)
        band = raster.GetRasterBand(1).ReadAsArray()
        pixel_count = raster.RasterXSize * raster.RasterYSize
        (histogram, _) = np.histogram(band, bins=np.arange(256))
        no_layover_shadow = 0.0
        true_layover = 0.0
        layover = 0.0
        true_shadow = 0.0
        shadow = 0.0
        pixel_count -= float(histogram[0])
        for i in range(len(histogram)):
            if i & 1:
                no_layover_shadow += float(histogram[i]) / pixel_count
            if i & 2:
                true_layover += float(histogram[i]) / pixel_count
            if i & 4:
                layover += float(histogram[i]) / pixel_count
            if i & 8:
                true_shadow += float(histogram[i]) / pixel_count
            if i & 16:
                shadow += float(histogram[i]) / pixel_count
        meta['ISO_RTC_noLayoverShadowPercentage'] = no_layover_shadow * 100.0
        meta['ISO_RTC_trueLayoverPercentage'] = true_layover * 100.0
        meta['ISO_RTC_layoverPercentage'] = layover * 100.0
        meta['ISO_RTC_trueShadowPercentage'] = true_shadow * 100.0
        meta['ISO_RTC_shadowPercentage'] = shadow * 100.0

    ## Look at the scattering area map
    scattering_area_file = log_file.replace('.log','_area.tif')
    if os.path.isfile(scattering_area_file):
        meta['ISO_RTC_scatteringAreaMap'] = \
            os.path.basename(scattering_area_file)
        raster = gdal.Open(product_file)
        band = raster.GetRasterBand(1)
        stats = band.GetStatistics(False, True)
        meta['ISO_RTC_scatteringMinValue'] = stats[0]
        meta['ISO_RTC_scatteringMaxValue'] = stats[1]
        meta['ISO_RTC_scatteringMeanValue'] = stats[2]
        meta['ISO_RTC_scatteringStandardDeviation'] = stats[3]

    ## Browse
    if polarization in ('DH','DV'):
        meta['ISO_RTC_browseImage'] = \
            os.path.basename(log_file.replace('.log','_rgb.png'))
        meta['ISO_RTC_KMLoverlay'] = \
            os.path.basename(log_file.replace('.log','_rgb.kmz'))
    else:
        meta['ISO_RTC_browseImage'] = \
            os.path.basename(log_file.replace('.log','.png'))
        meta['ISO_RTC_KMLoverlay'] = \
            os.path.basename(log_file.replace('.log','.kmz'))

    ## Extent
    (lon_min, lon_max, lat_min, lat_max) = get_latlon_extent(product_file)
    meta['ISO_RTC_westBoundLongitude'] = lon_min
    meta['ISO_RTC_eastBoundLongitude'] = lon_max
    meta['ISO_RTC_northBoundLatitude'] = lat_max
    meta['ISO_RTC_southBoundLatitude'] = lat_min

    return meta



#!/usr/bin/python3

"""Import packages"""
import argparse
from argparse import RawTextHelpFormatter
import os
import sys
import json
from metadata.asf_metadata import iso_template2lists, properties2values, \
  umm_dictionary_structure, meta_json_file


def sentinel_bursts_geojson2umm(excel_file, json_file, out_dir):
    """Convert burst metadata from GeoJSON to UMM-G format"""

    ### Check output directory
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    ### Get values for template placeholders
    with open(json_file, encoding='utf-8') as file:
        data = json.load(file)

    ### Loop through all bursts from the granule
    features = data['features']
    for feature in features:

        ### Read UMM-G template
        (umm_template, umm_params, umm_values) = \
            iso_template2lists(excel_file, 'UMM-G Structure')

        ### Extract UMM values
        properties = feature['properties']
        properties['insertDate'] = '{# UMM_insertDate #}'
        properties['hasLand'] = '{# UMM_hasLand #}'
        properties['crossesDateLine'] = '{# UMM_crossesDateLine #}'
        properties['landFraction'] = '{# UMM_landFraction #}'
        umm_dict = {}
        for key in properties.keys():
            umm_key = ('{# UMM_%s #}' % key)
            umm_dict[umm_key] = properties[key]
        properties = umm_dict
        umm_prod_values = properties2values(umm_values, properties)
        burst_id = properties['{# UMM_absoluteBurstID #}']

        ### Write UMM-G to file
        umm_file = f'{burst_id}-burst-umm.json'
        print(f'Writing UMM-G metadata to JSON file ({umm_file}) ...')
        umm_structure = \
            umm_dictionary_structure(umm_template, umm_params, umm_prod_values)
        meta_json_file(umm_structure, os.path.join(out_dir, umm_file))


if __name__ == '__main__':

    parser = argparse.ArgumentParser(prog='sentinel_bursts_geojson2umm.py',
        description='Convert Sentinel-1 bursts metadata from GeoJSON to ' \
        'UMM JSON format', formatter_class=RawTextHelpFormatter)
    parser.add_argument('excel_file', metavar='<Excel template>',
        help='name of the Excel template spreadsheet')
    parser.add_argument('json_file', metavar='<GeoJSON file>',
        help='name of the Sentinel-1 bursts metadata GeoJSON file')
    parser.add_argument('out_dir', metavar='<output directory>',
        help='name of the output directory')
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()

    sentinel_bursts_geojson2umm(args.excel_file, args.json_file, args.out_dir)
